//
// Simple NV_geometry_program4 example
//
// Demonstrates the use of use of assembly style geometry programs.
//  
// Author: Simon Green
// Email: sdkfeedback@nvidia.com
//
// Copyright (c) NVIDIA Corporation. All rights reserved.
////////////////////////////////////////////////////////////////////////////////

#define _CRT_SECURE_NO_DEPRECATE

#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <map>

#include <GL/glew.h>
#include <GL/glut.h>

//turn on dumping of program errors to stderr
#define NV_REPORT_COMPILE_ERRORS

#include "nvShaderUtils.h"
#include "nvGlutManipulators.h"
#include "nvMath.h"
#include <nvGlutWidgets.h>
#include <nvSDKPath.h>

//shader path relative to SDK root
#define SHADER_PATH "src/simple_geometry_program/shaders/"

using nv::vec3f;
using std::map;

nv::GlutExamine manipulator;

enum UIOption {
    OPTION_DISPLAY_WIREFRAME,
    OPTION_USE_GEOMETRY_PROGRAM,
    OPTION_ANIMATE,
    OPTION_COUNT
};
bool options[OPTION_COUNT];
map<char, UIOption> optionKeyMap;

nv::GlutUIContext ui;

// curve subdivisions
int segments = 20;

int width = 512;
int height = 512;

//Path finding code
nv::SDKPath sdkPath;

// vertex programs
static const char *xform_vprog_file = SHADER_PATH "xform.vp";

static const char *passthru_vprog_file = SHADER_PATH "passthrough.vp";

static const char *passthru_light_vprog_file = SHADER_PATH "passthrough_light.vp";


// geometry programs

// passthrough (tri to tri)
static const char *passthru_gprog_file = SHADER_PATH "passthrough.gp";

static const char *reflect_gprog_file = SHADER_PATH "reflect.gp";

// Cubic Bezier curve 
// creates a line strip from a line segment with adjacency
static const char *cubic_bezier_gprog_file = SHADER_PATH "cubic_bezier_curve.gp";

// Cubic B-Spline curve 
// creates a line strip from a line segement with adjacency
static const char *cubic_bspline_gprog_file = SHADER_PATH "cubic_bspline_curve.gp";

// Quadratic Bezier curve 
// takes a triangle as input, produces triangles
static const char *quadratic_bezier_gprog_file = SHADER_PATH "quadratic_bezier_curve.gp";

// turns a line into a wiggly line
// lots of customer demand for this one...
static const char *wiggle_gprog_file = SHADER_PATH "wiggle.gp";

// convert a point into a quad
static const char *pointsprite_gprog_file = SHADER_PATH "pointsprite.gp";

// convert a line into a thick line drawn as two triangles
static const char *thick_line_gprog_file = SHADER_PATH "thicklines.gp"; 

// subdivide a triangle once
static const char *subdivide_gprog_file = SHADER_PATH "subdivide.gp";

// subdivide a triangle recursively
static const char *subdivide_recursive_gprog_file = SHADER_PATH "recursive_subdivide.gp";

// create fins at each triangle edge
static const char *fins_gprog_file = SHADER_PATH "fins.gp";

// detect silhouette edges by comparing triangle normal
// with normals of adjacent triangles
// expects eye space position in vertex.texcoord
// 1-2-3
//  \|\|
//   0-4
//    \|
//     5
static const char *silhouette_gprog_file = SHADER_PATH "silhouette.gp";




enum ModelType {
    emtNone,
    emtLines,
    emtTriangles
};

enum ObjectType {
    eoNone,
    eoTeapot,
    eoSphere,
    eoTetrahedron
};

ObjectType object = eoTeapot;

struct Program {
    const char *name;
    const char *gprog_code, *vprog_code;
    GLenum prim; //type of primitive required
    ModelType modelType; //object to render by default
    ObjectType defaultObject;
    GLuint vprog, gprog;
};

Program prog[] = {
    { "point sprite",  pointsprite_gprog_file,  xform_vprog_file, GL_POINTS,  emtNone, eoNone, 0, 0 },
    { "thick line",  thick_line_gprog_file,  xform_vprog_file, GL_LINES,  emtLines, eoNone, 0, 0 },    
    { "cubic bezier curve",  cubic_bezier_gprog_file,  xform_vprog_file, GL_LINES_ADJACENCY_EXT,  emtNone, eoNone, 0, 0 },
    { "cubic b-spline curve",  cubic_bspline_gprog_file,  xform_vprog_file, GL_LINE_STRIP_ADJACENCY_EXT, emtNone, eoNone, 0, 0 },
    { "quadratic bezier curve",  quadratic_bezier_gprog_file,  xform_vprog_file, GL_TRIANGLES, emtTriangles, eoNone, 0, 0 },
    { "subdivide triangle",  subdivide_gprog_file,  xform_vprog_file, GL_TRIANGLES, emtTriangles, eoTetrahedron, 0, 0 },
    { "subdivide triangle recursive",  subdivide_recursive_gprog_file,  xform_vprog_file, GL_TRIANGLES, emtTriangles, eoTetrahedron, 0, 0 },
    { "fins",  fins_gprog_file,  passthru_vprog_file, GL_TRIANGLES, emtTriangles, eoTeapot, 0, 0 },    
    { "wiggle",  wiggle_gprog_file,  xform_vprog_file, GL_LINES, emtLines, eoSphere, 0, 0 },
    { "reflect",  reflect_gprog_file,  passthru_light_vprog_file, GL_TRIANGLES, emtTriangles, eoTeapot, 0, 0 },
    { "silhouette",  silhouette_gprog_file,  xform_vprog_file, GL_TRIANGLES_ADJACENCY_EXT, emtNone, eoNone, 0, 0 },
};

#define NPROGS sizeof(prog) / sizeof(Program)

Program *current_prog;
int current_prog_idx = 0;

////////////////////////////////////////////////////////////////////////////////
//
//  Functions
//
////////////////////////////////////////////////////////////////////////////////

//
//
//////////////////////////////////////////////////////////////////////
void init_opengl()
{
    glewInit();

    if (! glewIsSupported( "GL_VERSION_2_0 GL_ARB_vertex_program GL_ARB_fragment_program GL_NV_gpu_program4 ")) {
        printf( "Required extensions not supported.\n");
        printf( "This sample requires:\n");
        printf( "  OpenGL version 2.0 or later\n");
        printf( "  GL_NV_gpu_program4\n");
        printf( "  GL_ARB_vertex_program\n");
        printf( "  GL_ARB_fragment_program\n");
        exit(-1);
    }


	printf("%s\n\n", glGetString(GL_EXTENSIONS));

    // get geometry program limits
    GLint max_output_vertices, max_total_output_components;
    glGetProgramivARB(GL_GEOMETRY_PROGRAM_NV, GL_MAX_PROGRAM_OUTPUT_VERTICES_NV, &max_output_vertices);
    glGetProgramivARB(GL_GEOMETRY_PROGRAM_NV, GL_MAX_PROGRAM_TOTAL_OUTPUT_COMPONENTS_NV, &max_total_output_components);
    printf("max output vertices = %d\n", max_output_vertices);
    printf("max total output components = %d\n", max_total_output_components);

    glEnable(GL_DEPTH_TEST);
    glClearColor(0.8f, 0.8f, 0.8f, 1.0f);

    // load programs
    for(int i=0; i<NPROGS; i++) {
        std::string fullPath;
        printf( "Compiling %s\n", prog[i].name);

        if (sdkPath.getFilePath( prog[i].vprog_code, fullPath)) {
            prog[i].vprog = nv::CompileASMShaderFromFile(GL_VERTEX_PROGRAM_ARB, fullPath.c_str());
        }
        else {
            printf( "Failed to find vertex program file: %s\n", prog[i].vprog_code);
        }

        if (sdkPath.getFilePath( prog[i].gprog_code, fullPath)) {
            prog[i].gprog = nv::CompileASMShaderFromFile(GL_GEOMETRY_PROGRAM_NV, fullPath.c_str());
        }
        else {
            printf( "Failed to find geometry program file: %s\n", prog[i].gprog_code);
        }
    }

    current_prog = &prog[2];
    current_prog_idx = 2;
    object = current_prog->defaultObject;
}

//
//
//
//////////////////////////////////////////////////////////////////////
void drawObject()
{
    switch (object) {
        case eoTeapot:
            if (current_prog->modelType == emtTriangles)
                glutSolidTeapot(1.0);
            else
                glutWireTeapot(1.0);
            break;


        case eoSphere:
            if (current_prog->modelType == emtTriangles)
                glutSolidSphere(1.0, 10, 5);
            else
                glutWireSphere(1.0, 10, 5);
            break;

        case eoTetrahedron:
            if (current_prog->modelType == emtTriangles)
                glutSolidTetrahedron();
            else
                glutWireTetrahedron();
            
            break;
    }
}

//
//
//
/////////////////////////////////////////////////////////////////////
void drawSimplePrim()
{
    float v[][2] = {
        { 0.0, 0.0 },
        { 0.5, -0.2 },
        { 1.0, 0.4 },
        { 0.2, 0.8 },
        { -0.2, 1.0 },
        { -0.5, 0.0 },
        { -0.9, 0.3 }
    };
    int nverts = 7;

    switch(current_prog->prim) {
        case GL_POINTS:
            glBegin(GL_POINTS);
            glVertex2f(0.5, 0.5);
            glVertex2f(-0.2, 0.1);
            glVertex2f(0.7, -0.2);
            glEnd();
            break;

        case GL_LINES:
            glBegin(GL_LINES);
            for(int i=0; i<nverts-1; i++) {
                glVertex2fv(v[i]);
                glVertex2fv(v[i+1]);
            }
            glEnd();
            break;

        case GL_LINES_ADJACENCY_EXT:
            // 0 - - - 1----->2 - - - 3
            glBegin(GL_LINES_ADJACENCY_EXT);
            for(int i=0; i<nverts-1; i+=3) {
                    glVertex2fv(v[i]);
                    glVertex2fv(v[i+1]);
                    glVertex2fv(v[i+2]);
                    glVertex2fv(v[i+3]);
            }
            glEnd();
            if (!options[OPTION_USE_GEOMETRY_PROGRAM]) {
                // make adjacent lines visible
                glColor3f(1.0, 0.0, 0.0);
                glBegin(GL_LINES);
                for(int i=0; i<nverts-1; i+=3) {
                    glVertex2fv(v[i]);
                    glVertex2fv(v[i+1]);
                    glVertex2fv(v[i+2]);
                    glVertex2fv(v[i+3]);
                }
                glEnd();
            }
            break;

        case GL_LINE_STRIP_ADJACENCY_EXT:
            // 0 - - - 1--->2--->3--->4 - - - 5
            glBegin(GL_LINE_STRIP_ADJACENCY_EXT);
            for(int i=0; i<nverts; i++) {
                glVertex2fv(v[i]);
            }
            glEnd();
            if (!options[OPTION_USE_GEOMETRY_PROGRAM]) {
                // make adjacent lines visible
                glColor3f(1.0, 0.0, 0.0);
                glBegin(GL_LINES);
                glVertex2fv(v[0]);
                glVertex2fv(v[1]);
                glVertex2fv(v[nverts-2]);
                glVertex2fv(v[nverts-1]);
                glEnd();
            }
            break;

        case GL_TRIANGLES:
            glNormal3f(0.0f, 0.0f, 1.0f);
            glBegin(GL_TRIANGLES);
            glVertex3f(0.0f, 0.0f, 0.0f);
            glVertex3f(1.0f, 0.0f, 0.0f);
            glVertex3f(0.7f, 1.0f, 0.0f);

            glVertex3f(0.0f, 0.0f, 0.0f);
            glVertex3f(-1.0f, 0.0f, 0.0f);
            glVertex3f(-0.5f, -0.5f, 0.0f);
            glEnd();
            break;

        case GL_TRIANGLES_ADJACENCY_EXT:
            {
                // 1 2 3
                //   0 4
                //     5
                float v[][3] = {
                    0.0, 0.0, 0.0,
                    -1.0, 1.0, -1.0,
                    0.0, 1.0, 0.0,
                    1.0, 1.0, -1.0,
                    1.0, 0.0, 0.0,
                    1.0, -1.0, -1.0,
                };
                glBegin(GL_TRIANGLES_ADJACENCY_EXT);
                for(int i=0; i<6; i++) {
                    glVertex3fv(v[i]);
                }
                glEnd();
                if (!options[OPTION_USE_GEOMETRY_PROGRAM]) {
                    // draw adjacent tris
                    glColor3f(1.0, 0.0, 0.0);
                    glBegin(GL_TRIANGLES);
                    glVertex3fv(v[0]);
                    glVertex3fv(v[1]);
                    glVertex3fv(v[2]);

                    glVertex3fv(v[2]);
                    glVertex3fv(v[3]);
                    glVertex3fv(v[4]);

                    glVertex3fv(v[4]);
                    glVertex3fv(v[5]);
                    glVertex3fv(v[0]);
                    glEnd();
                }
            }
            break;


    }
}

//
//
//////////////////////////////////////////////////////////////////////
inline void updateButtonState( const nv::ButtonState &bs, nv::GlutManipulator &manip, int button) {
    int modMask = 0;

    if (bs.state & nv::ButtonFlags_Alt) modMask |= GLUT_ACTIVE_ALT;
    if (bs.state & nv::ButtonFlags_Shift) modMask |= GLUT_ACTIVE_SHIFT;
    if (bs.state & nv::ButtonFlags_Ctrl) modMask |= GLUT_ACTIVE_CTRL;

    if (bs.state & nv::ButtonFlags_End)
        manip.mouse( button, GLUT_UP, modMask, bs.cursor.x, height - bs.cursor.y);
    if (bs.state & nv::ButtonFlags_Begin)
        manip.mouse( button, GLUT_DOWN, modMask, bs.cursor.x, height - bs.cursor.y);
}

//
//
//////////////////////////////////////////////////////////////////////
void doUI()
{
    nv::Rect null;
    const char* programs[NPROGS];
    static const char* models[] = { "None", "Teapot", "Sphere", "Tetrahedron"};
    int currentModel = object;

    // get the programs presently available
    for (int ii = 0 ; ii < NPROGS; ii++) {
        programs[ii] = prog[ii].name;
    }

    glDisable( GL_DEPTH_TEST);

    ui.begin();

	ui.beginGroup( nv::GroupFlags_GrowDownFromLeft);

    ui.doLabel( null, "Simple Geometry Program");
    if ( ui.doComboBox( null, NPROGS, programs, &current_prog_idx)) {
        current_prog = &prog[current_prog_idx];
        object = current_prog->defaultObject;
    }
    ui.beginGroup( nv::GroupFlags_GrowLeftFromTop);
    ui.doLabel( null, "Model:");
    if (ui.doComboBox( null, 4, models, &currentModel)) {
        object = (ObjectType)currentModel;
    }
    ui.endGroup();

    ui.doCheckButton( null, "Wireframe", &options[OPTION_DISPLAY_WIREFRAME]);
    ui.doCheckButton( null, "Use geometry program", &options[OPTION_USE_GEOMETRY_PROGRAM]);

    // Pass non-ui mouse events to the manipulator
    if (!ui.isOnFocus()) {
        const nv::ButtonState &lbState = ui.getMouseState( 0);
        const nv::ButtonState &mbState = ui.getMouseState( 1);
        const nv::ButtonState &rbState =  ui.getMouseState( 2);

        manipulator.motion( ui.getCursorX(), height - ui.getCursorY());

        updateButtonState( lbState, manipulator, GLUT_LEFT_BUTTON);
        updateButtonState( mbState, manipulator, GLUT_MIDDLE_BUTTON);
        updateButtonState( rbState, manipulator, GLUT_RIGHT_BUTTON);
    }

	ui.endGroup();

    ui.end();

    glEnable( GL_DEPTH_TEST);
}

//
//  Rendering callback function
//
//////////////////////////////////////////////////////////////////////
void display()
{
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    manipulator.applyTransform();

    glBindProgramARB(GL_VERTEX_PROGRAM_ARB, current_prog->vprog);
    glBindProgramARB(GL_GEOMETRY_PROGRAM_NV, current_prog->gprog);

    glProgramLocalParameter4fARB(GL_GEOMETRY_PROGRAM_NV, 0, segments, 0, 0, 0);

    glPolygonMode(GL_FRONT_AND_BACK, options[OPTION_DISPLAY_WIREFRAME] ? GL_LINE : GL_FILL);

    if ( options[OPTION_USE_GEOMETRY_PROGRAM]) {
        glEnable(GL_VERTEX_PROGRAM_ARB);
        glEnable(GL_GEOMETRY_PROGRAM_NV);
    }

    glColor3f(0.0, 0.0, 0.0);
    glPointSize(2.0);

    if (current_prog->modelType == emtNone || object == eoNone)
        drawSimplePrim();
    else
        drawObject();

    glDisable(GL_VERTEX_PROGRAM_ARB);
    glDisable(GL_GEOMETRY_PROGRAM_NV);

    glutReportErrors();

    doUI();
    glutSwapBuffers();
}

//
//
//////////////////////////////////////////////////////////////////////
void idle()
{
    if (options[OPTION_ANIMATE])
        manipulator.idle();
    
    glutPostRedisplay();
}


//
//
//////////////////////////////////////////////////////////////////////
void key(unsigned char k, int x, int y)
{
    ui.keyboard( k, x, y);
    k = (unsigned char)tolower(k);

	//toggle the option, if we recognize it
    if ( optionKeyMap.find(k) != optionKeyMap.end())
        options[optionKeyMap[k]] = ! options[optionKeyMap[k]];

    switch(k) {
        case 27:
        case 'q':
            exit(0);
            break;


        case '=':
        case '+':
            if (segments < 40)segments++;
            break;

        case '-':
            if (segments > 1) segments--;
            break;
    }

    
	glutPostRedisplay();
}

//
//
//////////////////////////////////////////////////////////////////////
void resize(int w, int h)
{
    if (h == 0) h = 1;

    ui.reshape( w, h);

    glViewport(0, 0, w, h);
    
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    
    gluPerspective(60.0, (GLfloat)w/(GLfloat)h, 0.1, 100.0);
    
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    manipulator.reshape(w, h);
}

//
//
//////////////////////////////////////////////////////////////////////
void mouse(int button, int state, int x, int y)
{
    ui.mouse( button, state, glutGetModifiers(), x, y);
}

//
//
//////////////////////////////////////////////////////////////////////
void motion(int x, int y)
{
    ui.mouseMotion( x,y);
}

//
//
//////////////////////////////////////////////////////////////////////
void main_menu(int i)
{
  key((unsigned char) i, 0, 0);
}

//
//
//////////////////////////////////////////////////////////////////////
void prog_menu(int i)
{
    current_prog = &prog[i];
    object = current_prog->defaultObject;
    current_prog_idx = i;
}



//
//
//////////////////////////////////////////////////////////////////////
void init_menus()
{
    int prog_menu_id = glutCreateMenu(prog_menu);
    for(int i=0; i<NPROGS; i++) {
        glutAddMenuEntry( prog[i].name, i);
    }


    glutCreateMenu(main_menu);
    glutAddSubMenu("Program", prog_menu_id);
    glutAddMenuEntry("Increment segments [+]", '+');
    glutAddMenuEntry("Decrement segments [-]", '-');
    glutAddMenuEntry("Toggle geometry program [p]", 'p');
    glutAddMenuEntry("Toggle wireframe [w]", 'w');
    glutAddMenuEntry("Quit (esc)", '\033');
    glutAttachMenu(GLUT_RIGHT_BUTTON);
}

//
//
//////////////////////////////////////////////////////////////////////
int main(int argc, char **argv)
{
	glutInit(&argc, argv);
	glutInitWindowSize(512, 512);
	glutInitDisplayMode(GLUT_DOUBLE | GLUT_DEPTH | GLUT_RGB);
	glutCreateWindow("simple_geometry_program");

	init_opengl();

    // configure the manipulator for single button control
    manipulator.setDollyActivate( GLUT_LEFT_BUTTON, GLUT_ACTIVE_SHIFT);
    manipulator.setPanActivate( GLUT_LEFT_BUTTON, GLUT_ACTIVE_CTRL);

    manipulator.setDollyPosition( -2.0f);

	glutDisplayFunc(display);
    glutMouseFunc(mouse);
    glutMotionFunc(motion);
    glutPassiveMotionFunc(motion);
    glutIdleFunc(idle);
    glutKeyboardFunc(key);
    glutReshapeFunc(resize);
    init_menus();

    
    //configure the options
    optionKeyMap['w'] = OPTION_DISPLAY_WIREFRAME;
    options[OPTION_DISPLAY_WIREFRAME] = false;

    optionKeyMap['p'] = OPTION_USE_GEOMETRY_PROGRAM;
    options[OPTION_USE_GEOMETRY_PROGRAM] = true;
    
    optionKeyMap[' '] = OPTION_ANIMATE;
    options[OPTION_ANIMATE] = false;

    //print the help info
    printf( "Simple_geometry_program - sample showing the usage of GL_NV_geometry_program4\n");
    printf( "  Commands:\n");
    printf( "    q / [ESC] - Quit the application\n");
    printf( "    [SPACE]   - Toggle continuous animation\n");
    printf( "    p         - Toggle using geometry and vertex programs\n\n");
    printf( "    w         - Toggle displaying wireframe\n");
    printf( "    +         - Increase the number of segments used for tesselating curves\n");
    printf( "    -         - Decrease the number of segments used for tesselating curves\n");
    printf( "    [mouse]   - Left button - rotate, Left + shift - dolly, Left + ctrl - pan\n");
    printf( "    [mouse]   - Decrease the number of segments used for tesselating curves\n\n");

	glutMainLoop();

	return 0;
}